﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Web;
using System.IO;
using System.Collections.Specialized;

namespace cosc427_design.ServerRequests
{
    class BasicSender : HttpPOSTRequestSender
    {
        // Server URL to POST to
        private string postUrl;

        private WebRequest serverRequest = null;

        /// <summary>
        /// Creates a new instance.
        /// </summary>
        /// <param name="url">The URL the request will be sent to.</param>
        /// <exception cref="System.ArgumentNullException"> When a null url is supplied. </exception>
        public BasicSender(string url)
        {
            if (url == null)
            {
                throw new ArgumentNullException("All method arguments must be valid non-null values.");
            }
            this.postUrl = url;
        }

        /// <summary>
        /// Sends the Http POST request to the server.
        /// </summary>
        /// <param name="requestObj">Object specifying the request.</param>
        /// <returns>True if the request was successfully sent.</returns>
        /// 
        public override bool Send(System.Collections.Specialized.NameValueCollection requestData)
        {
            try
            {
                this.serverRequest = WebRequest.Create(this.postUrl);
                this.serverRequest.ContentType = "application/x-www-form-urlencoded";
                this.serverRequest.Method = "POST";

                // Prepare POST data
                string postString = GeneratePOSTString(requestData);
                byte[] paramBytes = Encoding.UTF8.GetBytes(postString);
                
                // Send request
                if (paramBytes == null) throw new ArgumentException("Cannot send a request with no POST data.");
                this.serverRequest.ContentLength = paramBytes.Length;
                this.serverRequest.Timeout = 10000;
                using(Stream requestStream = this.serverRequest.GetRequestStream())
                {
                    requestStream.Write(paramBytes, 0, paramBytes.Length);
                    requestStream.Close();
                }
            }
            catch (WebException)
            {
                // Network unavailable while trying to get remote stream
                return false;
            }
            catch (IOException)
            {
                // Network becomes unavailable during writing to remote byte stream (sending)
                return false;
            }
            return true;
        }

        /// <summary>
        /// Receives the response resulting from the POST request made to the server.
        /// This method can only be called once per (and after) a successful call to Send().
        /// </summary>
        /// <returns>The response content.</returns>
        /// <exception cref=""
        public override byte[] ReceiveResponse()
        {
            if (this.serverRequest == null) 
                throw new InvalidOperationException("A call to ReceiveResponse() is only valid after a successful call to SendRequest().");

            byte[] responseContent = null;
            try
            {
                using (WebResponse webResponse = this.serverRequest.GetResponse())
                {
                    if (webResponse == null) return null;
                    Stream responseStream = webResponse.GetResponseStream();
                    responseContent = ReadFully(responseStream, -1);
                    responseStream.Close();
                    webResponse.Close();

                    // Every call to Send can only correspond to a single call to Receive
                    this.serverRequest = null;
                }
            }
            catch (IOException)
            {
                // Any IO problem due to web stream interruption
                throw new WebException();
            }
            return responseContent;
        }

        /// <summary>
        /// Generates the POST string for the request.
        /// </summary>
        /// <returns>The POST string for the request.</returns>
        private string GeneratePOSTString(NameValueCollection postData)
        {
            if (postData.Count == 0) throw new ArgumentException("There are no entries in the POST string collection.");

            string firstKey = postData.AllKeys[0];
            string firstValue = postData[firstKey];

            StringBuilder postStrBuilder = new StringBuilder();
            // First parameter in POST string is only that need not be preceeded by "&" 
            postStrBuilder.Append(firstKey);
            postStrBuilder.Append("=");
            postStrBuilder.Append(HttpUtility.UrlEncode(firstValue));
            
            for (int i = 1; i <  postData.AllKeys.Length; i++)
            {
                string currentKey = postData.AllKeys[i];
                string currentValue = postData[currentKey];

                // Add parameter and value
                postStrBuilder.Append("&");
                postStrBuilder.Append(currentKey);
                postStrBuilder.Append("=");
                // UrlEncode() to escape special symbols that are potentially present
                postStrBuilder.Append(HttpUtility.UrlEncode(currentValue));
            }
            return postStrBuilder.ToString();
        }

        /// <summary>
        /// Borrowed from: http://www.yoda.arachsys.com/csharp/readbinary.html
        /// 
        /// Reads data from a stream until the end is reached. The
        /// data is returned as a byte array. An IOException is
        /// thrown if any of the underlying IO calls fail.
        /// </summary>
        /// <param name="stream">The stream to read data from</param>
        /// <param name="initialLength">The initial buffer length</param>
        private static byte[] ReadFully(Stream stream, int initialLength)
        {
            // If we've been passed an unhelpful initial length, just
            // use 32K.
            if (initialLength < 1)
            {
                initialLength = 32768;
            }

            byte[] buffer = new byte[initialLength];
            int read = 0;

            int chunk;
            while ((chunk = stream.Read(buffer, read, buffer.Length - read)) > 0)
            {
                read += chunk;

                // If we've reached the end of our buffer, check to see if there's
                // any more information
                if (read == buffer.Length)
                {
                    int nextByte = stream.ReadByte();

                    // End of stream? If so, we're done
                    if (nextByte == -1)
                    {
                        return buffer;
                    }

                    // Nope. Resize the buffer, put in the byte we've just
                    // read, and continue
                    byte[] newBuffer = new byte[buffer.Length * 2];
                    Array.Copy(buffer, newBuffer, buffer.Length);
                    newBuffer[read] = (byte)nextByte;
                    buffer = newBuffer;
                    read++;
                }
            }
            // Buffer is now too big. Shrink it.
            byte[] ret = new byte[read];
            Array.Copy(buffer, ret, read);
            return ret;
        }
    }
}
