﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;
using System.Collections.Specialized;

namespace cosc427_design.ServerRequests
{
    public class Request
    {
        private const string REQ_SUCCESS = "1";
        private const string REQ_FAILURE = "0";

        private NameValueCollection data;
        private HttpPOSTRequestSender sender;
        private bool sent = false;
        private string response;

        public Request(HttpPOSTRequestSender sender)
        {
            this.sender = sender;
        }

        public HttpPOSTRequestSender Sender
        {
            get
            {
                return this.sender;
            }
            set
            {
                this.sender = value;
            }
        }

        public NameValueCollection Data
        {
            get
            {
                return this.data;
            }
        }

        public virtual void AddParameter(string paramName, string paramValue)
        {
            if (this.data == null)
            {
                this.data = new NameValueCollection();
            }

            this.data.Add(paramName, paramValue);
        }

        /// <summary>
        /// Performs post execution tasks. Stub for child classes that want to further deal with server.
        /// </summary>
        public virtual void ConfirmOutcome()
        {
            // Not implemented. Left for sub classes.
        }

        public virtual void PopulateData(RegistrationData destination)
        {
            // Not implemented. Left for sub classes.
        }

        /// <summary>
        /// Execute the request to the server.
        /// <returns> True if the request executed successfully.</returns>
        /// </summary>
        public virtual bool Execute()
        {
            bool executionSuccessful = Send();
            executionSuccessful &= RequestIsSuccessful();
            return executionSuccessful;
        }

        /// <summary>
        /// Sends the current request to its destination.
        /// </summary>
        /// <returns>True if the request was successfully sent.</returns>
        private bool Send()
        {
            this.response = null;
            this.sent = this.sender.Send(this.data);
            return this.sent;
        }

        /// <summary>
        /// Receives the content intended as a response to the sent request. 
        /// </summary>
        /// <returns>An XML string with the response content received. </returns>
        /// <exception cref="WebException">Occurs when the response cannot be obtained because of connection problems.</exception>
        protected string GetResponse()
        {
            if (!this.sent)
            {
                throw new InvalidOperationException("GetResponse() may only be called after a successful call to Send().");
            }

            if (this.response == null)
            {
                MemoryStream responseStream = new MemoryStream(this.sender.ReceiveResponse());
                StreamReader responseReader = new StreamReader(responseStream);
                this.response = responseReader.ReadToEnd();
                responseReader.Close();
            }

            return this.response;
        }

        /// <summary>
        /// Specifies whether the sent request executed successfully at the registration server.
        /// </summary>
        /// <returns>True if the request was successful.</returns>
        /// <exception cref="WebException">Thrown when the server response cannot be obtained due to connection problems.</exception>
        public virtual bool RequestIsSuccessful()
        {
            bool requestSuccess = false;
            string responseCode;
            try
            {
                XmlDocument responseDoc = new XmlDocument();
                string xmlString = this.GetResponse();
                responseDoc.LoadXml(xmlString);

                // Read in the response code and check it is as expected
                XmlNode responseElement = responseDoc.GetElementsByTagName("Response").Item(0);
                responseCode = responseElement.Attributes["code"].Value;

                if (responseCode.Equals(Request.REQ_SUCCESS))
                {
                    requestSuccess = true;
                }
            }
            catch (XmlException)
            {
                // we have received XML we don't understand
                requestSuccess = false;
            }

            return requestSuccess;
        }






    }
}
